﻿using System;
using System.Collections.Generic;
using System.Linq;
using UnityEngine;

namespace JudgeModule
{ 
    public class NoteManager
    {
        private float baseTime;
        private bool IsInitTime;
    
        private GameObject offset;
        private GameObject noteobj;
        private GameObject deactives;
        private GameObject appear;
        private GameObject disappear;
    
        private Vector3 initialPos;
        private float BPM;
        private Func<Note> GetLastNote;
    
        private float MsPerBeat
        {
            get
            {
                return 60 * 1000f / BPM;
            }
        }
    
        private float ScrollSpeed
        {
            get
            {
                return MsPerBeat / 1000f;
            }
        }
        
        public NoteManager(Dictionary<string, GameObject> gameObjects, float bpm, Func<Note> getlastnote)
        {
            offset    = gameObjects["offset"];
            noteobj   = gameObjects["noteobj"];
            deactives = gameObjects["deactives"];
            appear    = gameObjects["appear"];
            disappear = gameObjects["disappear"];
    
            initialPos = offset.transform.position;
            BPM = bpm;
            GetLastNote = getlastnote;
        }
    
        public void Controll(float current)
        {
            if (!IsInitTime)
            {
                baseTime = current;
                IsInitTime = true;
            }
    
            float timing = current - baseTime;
    
            offset.transform.position = Vector3.left * timing * ScrollSpeed;
    
            RelocateControllers(timing);
        }
    
        private void RelocateControllers(float timing)
        {
            DeactiveUndecided();
            SwitchStandBy();
            AssignNotes(timing);
        }
    
        private void DeactiveUndecided()
        {
            noteobj.transform
                   .Cast<Transform>()
                   .Select(x => x.GetComponent<Controller>())
                   .Where(c => initialPos.x - c.transform.position.x
                        - (c.Instance.EndTiming - c.Instance.StartTiming)
                        > Judge.BAD.ButtonTimingRange)
                   .ToList()
                   .ForEach(x => x.transform.SetParent(deactives.transform));
        }
    
        private void SwitchStandBy()
        {
            deactives.transform
                     .Cast<Transform>()
                     .Select(x => x.GetComponent<Controller>())
                     .Where(c => c.transform.position.x <= disappear.transform.position.x)
                     .ToList()
                     .ForEach(x =>
                     {
                         x.transform.position = appear.transform.position;
                         x.transform.SetParent(appear.transform);
                         x.Instance = null;
                         x.gameObject.SetActive(false);
                     });
        }
    
        private void AssignNotes(float timing)
        {
            appear.transform
                  .Cast<Transform>()
                  .Select(x => x.GetComponent<Controller>())
                  .ToList()
                  .ForEach(c =>
                  {
                      var note = GetLastNote();
                      if (note != null)
                      {
                          c.Instance = note;
                          c.Instance.Component = c;
                          c.transform.position = initialPos + Vector3.right * ((c.Instance.StartTiming - timing) * ScrollSpeed);
                          c.transform.SetParent(noteobj.transform);
                          c.gameObject.SetActive(true);
                      }
                  });
        }
    }
}